#include <math.h>
#include "linecyln.h"

// General quadric equation is X^T A X + B^T X + C = 0 where A is a 3x3
// symmetric matrix, B is a 3x1 vector, and C is a scalar.  X is a 3x1
// vector of unknowns.  Factor A = R^T D R using an eigendecomposition
// where R is an orthonormal matrix and D = diag(d0,d1,d2) is a diagonal
// matrix.  Define E = RB and Y = RX.  The quadric becomes
//
//   Y^T D Y + E^T Y + C = 0
//
// Let Y = (y0,y1,y2) and E = (e0,e1,e2).  If d0 = d1 = 1 and d2 = 0 and
// e2 = 0, the equation describes a cylinder.  The last row of R is the
// cylinder axis (last row corresponds to eigenvalue d2 = 0).  Completing
// the square yields
//
//   (y0+0.5*e0)^2 + (y1+0.5*e1)^2 = 0.25*(e0^2+e1^2) - C.
//
// If the right-hand side is positive, then the cylinder squared radius
// is the right-hand side.  A point on the cylinder axis is -(e0,e1,0)/2
// = -E/2, where the components are in the Y coordinate system.  In the
// original X coordinates, the point is P = -R^T E/2.  The direction of
// the cylinder axis is (0,0,1) in the Y coordinate system.  In the X
// coordinate system, the axis is F = R^T (0,0,1) [the last row of R).
//
//
// The reverse problem is to specify the cylinder axis P+t*F and radius
// r, then compute the corresponding quadric equation X^T A X + B^T X + C
// = 0.  Choose vectors U and V so that U, V, and F form an orthonormal
// set.  Transpose of orthogonal matrix is R^T = [U V F] where U, V, and
// F are the columns of the matrix.  Diagonal matrix is D = diag(1,1,0).
// E = -2*R*P implies B = R^T E = -2*P.  Matrix A = R^T D R.  Finally,
// r^2 = 0.25*(e0^2+e1^2)-C implies C = 0.25*(e0^2+e1^2)-r^2.
//
//
// Given a line X(t) = Q+t*G and quadric equation X^T A X + B^T X + C = 0,
// plug in X(t) to get
//
//   (G^T A G) t^2 + (B^T G + 2 Q^T A G) t + (Q^T A Q + B^T Q + C) = 0.
//
// or c0+c1*t+c2*t^2.  Solve for t.  If two distinct real roots, line
// intersects cylinder in two points.  If one repeated real root, line
// is tangent to cylinder.  If no real roots, line and cylinder do not
// intersect.

//---------------------------------------------------------------------------
void IntersectLineCylinder (const Line3& line, const Cylinder3& cylinder,
    int& numIntersections, Point3 intersection[2])
{
    // choose U, V so that U,V,F is orthonormal set
    Point3 U, V, F = cylinder.axis.direction;
    float length, invLength, prod;
    if ( fabs(F.x) > fabs(F.y) && fabs(F.x) > fabs(F.z) )
    {
        length = float(sqrt(F.x*F.x+F.y*F.y));
        invLength = 1.0f/length;
        U.x = -F.y*invLength;
        U.y = +F.x*invLength;
        U.z = 0.0f;
        prod = -F.z*invLength;
        V.x = F.x*prod;
        V.y = F.y*prod;
        V.z = length;
    }
    else
    {
        length = float(sqrt(F.y*F.y+F.z*F.z));
        invLength = 1.0f/length;
        U.x = 0.0f;
        U.y = -F.z*invLength;
        U.z = +F.y*invLength;
        prod = -F.x*invLength;
        V.x = length;
        V.y = F.y*prod;
        V.z = F.z*prod;
    }

    // orthonormal matrix
    float R[3][3] = 
    {
        { U.x, U.y, U.z },
        { V.x, V.y, V.z },
        { F.x, F.y, F.z }
    };

    // matrix A
    float A[3][3] =
    {
        {
          R[0][0]*R[0][0]+R[1][0]*R[1][0],
          R[0][0]*R[0][1]+R[1][0]*R[1][1],
          R[0][0]*R[0][2]+R[1][0]*R[1][2]
        },
        {
          R[0][1]*R[0][0]+R[1][1]*R[1][0],
          R[0][1]*R[0][1]+R[1][1]*R[1][1],
          R[0][1]*R[0][2]+R[1][1]*R[1][2]
        },
        {
          R[0][2]*R[0][0]+R[1][2]*R[1][0],
          R[0][2]*R[0][1]+R[1][2]*R[1][1],
          R[0][2]*R[0][2]+R[1][2]*R[1][2]
        }
    };

    // vector B
    Point3 P = cylinder.axis.origin;
    Point3 B = { -2.0f*P.x, -2.0f*P.y, -2.0f*P.z };

    // constant C
    float e0 = -2.0f*(R[0][0]*P.x+R[0][1]*P.y+R[0][2]*P.z);
    float e1 = -2.0f*(R[1][0]*P.x+R[1][1]*P.y+R[1][2]*P.z);
    float C = 0.25f*(e0*e0+e1*e1) - cylinder.radius*cylinder.radius;

    // line
    Point3 Q = line.origin;
    Point3 G = line.direction;

    // compute A*G
    Point3 AG =
    {
        A[0][0]*G.x+A[0][1]*G.y+A[0][2]*G.z,
        A[1][0]*G.x+A[1][1]*G.y+A[1][2]*G.z,
        A[2][0]*G.x+A[2][1]*G.y+A[2][2]*G.z
    };

    // compute A*Q
    Point3 AQ =
    {
        A[0][0]*Q.x+A[0][1]*Q.y+A[0][2]*Q.z,
        A[1][0]*Q.x+A[1][1]*Q.y+A[1][2]*Q.z,
        A[2][0]*Q.x+A[2][1]*Q.y+A[2][2]*Q.z
    };

    // compute quadratic equation c0+c1*t+c2*t^2 = 0
    float c2 = G.x*AG.x+G.y*AG.y+G.z*AG.z;
    float c1 = B.x*G.x+B.y*G.y+B.z*G.z+2.0f*(Q.x*AG.x+Q.y*AG.y+Q.z*AG.z);
    float c0 = Q.x*AQ.x+Q.y*AQ.y+Q.z*AQ.z+B.x*Q.x+B.y*Q.y+B.z*Q.z+C;

    // solve for intersections
    const float epsilon = 1e-06f;
    float discr = c1*c1-4.0f*c0*c2;
    if ( discr > epsilon )
    {
        numIntersections = 2;
        discr = float(sqrt(discr));
        float invC2 = 1.0f/c2;
        float root0 = -0.5f*(c1+discr)*invC2;
        float root1 = -0.5f*(c1-discr)*invC2;
        intersection[0].x = Q.x+root0*G.x;
        intersection[0].y = Q.y+root0*G.y;
        intersection[0].z = Q.z+root0*G.z;
        intersection[1].x = Q.x+root1*G.x;
        intersection[1].y = Q.y+root1*G.y;
        intersection[1].z = Q.z+root1*G.z;
    }
    else if ( discr < -epsilon )
    {
        numIntersections = 0;
    }
    else
    {
        numIntersections = 1;
        float root = -0.5f*c1/c2;
        intersection[0].x = Q.x+root*G.x;
        intersection[0].y = Q.y+root*G.y;
        intersection[0].z = Q.z+root*G.z;
    }
}
//---------------------------------------------------------------------------

#ifdef LINECYLN_TEST

void main ()
{
    Line3 line;
    line.origin.x = 0.0f;
    line.origin.y = 0.0f;
    line.origin.z = 0.0f;
    line.direction.x = 1.0f;
    line.direction.y = 0.0f;
    line.direction.z = 0.0f;

    Cylinder3 cylinder;
    cylinder.axis.origin.x = 0.0f;
    cylinder.axis.origin.y = 0.0f;
    cylinder.axis.origin.z = 0.0f;
    cylinder.axis.direction.x = 0.0f;
    cylinder.axis.direction.y = 0.0f;
    cylinder.axis.direction.z = 1.0f;
    cylinder.radius = 1.0f;

    int numIntersections;
    Point3 intersection[2];
    IntersectLineCylinder(line,cylinder,numIntersections,intersection);
}

#endif
